/*  FrvTF.c by  F. Marion and B. Mours    May 01, 2002           */

#include "FrvTF.h"
#include "FrvMath.h"
#include <string.h>
#include <math.h>

/*-------------------------------------------------------------- FrvTFError--*/
void FrvTFError(FrvTF *tf)
/*---------------------------------------------------------------------------*/
/*   Returns: 0 if OK or an error code in case of problem.                   */
/*---------------------------------------------------------------------------*/
{double dnu, *errorM, *errorP, *coherence, *modulus, coefA, coefB, term;
 int nData, i;

 if(tf == NULL) return;
 if(tf->optionC != FR_YES) return; 

                /*----- create output structures if they are not available---*/

 nData = tf->modulus->nData;

 if(tf->errorM == NULL)
   { dnu   = tf->modulus->dx[0];
     tf->errorM = FrVectNew1D("modulusErr",FR_VECT_8R, nData, dnu, 
                                       "Frequency [Hz]", "modulus Error");
     tf->errorP  = FrVectNew1D("phaseErr",FR_VECT_8R, nData, dnu,
                                       "Frequency [Hz]", "phase error");
    if(tf->errorM == NULL) return;
    if(tf->errorP == NULL) return;}

               /*------------------------------------ compute the errors ---*/

 errorM    = tf->errorM->dataD;
 errorP    = tf->errorP->dataD;    
 coherence = tf->coherence->dataD;
 modulus   = tf->modulus->dataD;

 coefA = .85/sqrt(tf->nCall);
 coefB = .88/sqrt(tf->nCall);

 for(i=0; i<nData; i++)  
   {term = sqrt((1.-coherence[i])/coherence[i]);
    errorM[i] = coefA*term*modulus[i];
    errorP[i] = coefB*term;}

 return;}

/*-------------------------------------------------------------- FrvTFFree--*/
void FrvTFFree(FrvTF *tf)
/*---------------------------------------------------------------------------*/
/*  Free the TF object.                                                      */
/*---------------------------------------------------------------------------*/
{
 if(tf == NULL) return;

 FrvBufFree(tf->bufferS);
 FrvBufFree(tf->bufferR);

 FrvRFFTFree(tf->fftS);
 FrvRFFTFree(tf->fftR);

 FrVectFree(tf->output);
 FrVectFree(tf->correlation);
 FrVectFree(tf->modulus);
 FrVectFree(tf->phase);

 if(tf->coherence != NULL) FrVectFree(tf->coherence);
    
 free(tf);
                        
 return;
}
/*---------------------------------------------------------------- FrvTFIni--*/
int FrvTFInit(FrvTF *tf,
              FrVect *signal, 
              FrVect *ref)
/*---------------------------------------------------------------------------*/
/*   Returns: 0 if OK or an error code in case of problem.                   */
/*---------------------------------------------------------------------------*/
{char name[256];
 double dnu;
 int nData, step,i ;


 if(tf     == NULL) return(1);
 if(signal == NULL) return(1);
 if(ref    == NULL) return(1);       

            /*------- create FFT buffer object -----*/

 if(tf->outSize < 1)  tf->outSize = signal->nData;
 step = tf->outSize/2;

 tf->bufferS = FrvBufNew(tf->outSize, step, FR_VECT_8R, tf->decimate, 0);
 tf->bufferR = FrvBufNew(tf->outSize, step, FR_VECT_8R, tf->decimate, 0);

 if(tf->bufferS == NULL) return(2);
 if(tf->bufferR == NULL) return(2);

            /*------- create FFT object -----*/

 if((tf->fftS = FrvRFFTNew("HA",0,0)) == NULL) return(2);
 if((tf->fftR = FrvRFFTNew("HA",0,0)) == NULL) return(2);
 FrvRFFTSetDecay(tf->fftS, tf->decay);
 FrvRFFTSetDecay(tf->fftR, tf->decay);

                 /*--------- create output structures ----------------------*/

 sprintf(name,"TF(%s/%s)",signal->name,ref->name);
 nData = tf->outSize/2;
 dnu = 1./(2*nData*tf->decimate*signal->dx[0]);

 tf->output  = FrVectNew1D(name, FR_VECT_C16,nData,dnu,"Frequency [Hz]","T.F");

 tf->correlation = FrVectNew1D("correlation", FR_VECT_C16,nData,dnu,
                                         "Frequency [Hz]","T.F correlation");

 tf->modulus = FrVectNew1D("modulus",FR_VECT_8R, nData, dnu, "Frequency [Hz]",
                           "modulus");
 tf->phase   = FrVectNew1D("phase",FR_VECT_8R, nData, dnu, "Frequency [Hz]",
                           "phase [rad]");
 if(tf->output      == NULL) return(1);
 if(tf->correlation == NULL) return(1);
 if(tf->modulus     == NULL) return(1);
 if(tf->phase       == NULL) return(1);
    
 if(tf->optionC == FR_YES) 
   {tf->coherence = FrVectNew1D("coherence",FR_VECT_8R, nData, dnu,
                                "Frequency[Hz]", "coherence");
    if(tf->coherence == NULL) return(1);}

 for(i=0; i<2*tf->correlation->nData; i++)  {tf->correlation->dataD[i] = 0.;}
 for(i=0; i<tf->phase->nData; i++)  {tf->phase->dataD[i] = 0.;}
    
         /*-------------------- adjust bin origin ---------------------------*/

 tf->output     ->startX[0] = 0.5*dnu;
 tf->correlation->startX[0] = 0.5*dnu;
 tf->modulus    ->startX[0] = 0.5*dnu;
 tf->phase      ->startX[0] = 0.5*dnu;

 return(0);
}
/*-------------------------------------------------------------- FrvTFNew -*/
FrvTF* FrvTFNew(char* option, int outSize, int decimate)
/*---------------------------------------------------------------------------*/
{FrvTF *tf;

 tf = (FrvTF *) (calloc(1, sizeof(FrvTF)));
 if(tf == NULL) return(NULL);

 tf->optionC = FR_NO;
 tf->optionE = FR_NO;
 if(option != NULL)
   {if(strchr(option,'C') != NULL) tf->optionC = FR_YES;
    if(strchr(option,'E') != NULL) tf->optionE = FR_YES;}

 tf->outSize   = outSize;
  
 if(decimate < 1)   decimate = 1;
 tf->decimate  = decimate;
 tf->decay   = .999;
 tf->nCall   = 0;
                        
 return(tf);
}
/*-------------------------------------------------------------- FrvTFProc--*/
int FrvTFProc(FrvTF *tf, 
              FrVect *signal, 
              FrVect *ref)
/*---------------------------------------------------------------------------*/
/*   Returns: 0 if OK or an error code in case of problem                    */
/*---------------------------------------------------------------------------*/
{
 int i;
 double  decay, modNum,
         *output, *correlation, *modulus, *coherence, *signalA, *noiseA;

 if(tf     == NULL)  return(1);
 if(signal == NULL)  return(1);
 if(ref    == NULL)  return(1);
 if(signal->nData != ref->nData) return(2);

          /*------------ initialized tf object if needed ------------------*/

 if(tf->output == NULL) {if(FrvTFInit(tf,signal, ref) != 0) return(3);}

          /*---------------------- feed data --------------------------------*/

 FrvBufFeed(tf->bufferS,signal);     
 FrvBufFeed(tf->bufferR,ref);     
 
          /*---------------------------- now loop over the sub samples ------*/

 while(FrvBufGetNext(tf->bufferS) == 0)

   {if(FrvBufGetNext(tf->bufferR) != 0) return(4);

    tf->nCall++;
    decay = 1. - 1./tf->nCall;
    if(tf->decay < decay) decay = tf->decay;

              /*------------------------------- compute FFT ----------------*/

    FrvRFFTFor(tf->fftS, tf->bufferS->output);
    FrvRFFTFor(tf->fftR, tf->bufferR->output);

              /*------------------ compute the current TF correlation--------*/
              /*------------- W: we make temporary use of tf->output---------*/

    FrvMultConj(tf->fftS->output,tf->fftR->output,tf->output,NULL);

              /*--------------------- compute the mean correlation-----------*/

    output    = tf->output->dataD;
    correlation = tf->correlation->dataD;

    for(i=0; i<2*tf->output->nData; i++) 
        {correlation[i] = decay*correlation[i] +(1.-decay)*output[i];}

              /*-------- -------- compute the mean modulus and coherance------*/

    modulus   = tf->modulus->dataD;
    signalA   = tf->fftS->amplitudeA->dataD;
    noiseA    = tf->fftR->amplitudeA->dataD;
    if(tf->optionC == FR_YES) 
         {coherence = tf->coherence->dataD;}
    else {coherence = NULL;}

    for(i=0; i<tf->output->nData; i++) 
        {modNum = sqrt(correlation[2*i]  *correlation[2*i]  +
                       correlation[2*i+1]*correlation[2*i+1]);
         modulus[i] = modNum/(noiseA[i]*noiseA[i]);
         if(tf->optionC == FR_YES) 
                coherence[i] = modulus[i]*modNum/(signalA[i]*signalA[i]);}

              /*------------------------ compute the mean phase-------------*/

    FrvPhase(tf->correlation,tf->phase,NULL);
 
              /*--------------------------- compute the current TF ----------*/

    FrvDivide(tf->fftS->output,tf->fftR->output, tf->output, NULL);}


               /*------------------- compute the TF error if requested-------*/

 if(tf->optionE == FR_YES) FrvTFError(tf); 
 
                     /*------- Set start time for all the series defined --- */

 tf->output->GTimeS = signal->GTimeS;
 tf->output->GTimeN = signal->GTimeN;
 tf->modulus->GTimeS = signal->GTimeS;
 tf->modulus->GTimeN = signal->GTimeN;
 if(tf->coherence != NULL) 
   {tf->coherence->GTimeS = signal->GTimeS;
    tf->coherence->GTimeN = signal->GTimeN;}
 if(tf->errorM != NULL) 
   {tf->errorM->GTimeS = signal->GTimeS;
    tf->errorM->GTimeN = signal->GTimeN;}
 if(tf->errorP != NULL) 
   {tf->errorP->GTimeS = signal->GTimeS;
    tf->errorP->GTimeN = signal->GTimeN;}

 return(0);
}

/*---------------------------------------------------------- FrvTFSetDecay --*/
void FrvTFSetDecay(FrvTF* tf, double decay)
/*---------------------------------------------------------------------------*/
/*   Change the decay value for the mean amplitude computation               */
/*---------------------------------------------------------------------------*/
{
 if(tf  == NULL)  return;

 if(decay < 0.) decay = 0.;
 if(decay > 1.) decay = 1.;
 tf->decay = decay;
    
 FrvRFFTSetDecay(tf->fftS, tf->decay);
 FrvRFFTSetDecay(tf->fftR, tf->decay);

 return;}


